// Copyright 2009 Eric Estabrooks. All rights reserved.
// Use of this source code is governed by a BSD-style
// license that can be found in the LICENSE file.

// This package provides an irc client object
// it handles communication with server
// mostly just passes stuff through
// sometimes it eats things it shouldn't
package ircclient

import "os"
import "net"
import "fmt"
import "strings"
import "regexp"
import "crypto/tls"
import "time"
import "bytes"

var ping *regexp.Regexp = regexp.MustCompile(".*PING.* :(.*)");

type IRCClient struct {
	tcon        *tls.Conn;
	ncon        net.Conn;
	buff        *bytes.Buffer;
	CanRead     chan bool;
	control     chan bool;
	name        string;
	Nick        string;
	Channel     string;
	valid       bool;
}

func New() *IRCClient {
	readChan := make(chan bool, 32);
	conChan := make(chan bool);

	return &IRCClient{ CanRead: readChan, control:conChan };
}


func (client *IRCClient) Read(b []byte) (n int, err os.Error) {
	if client != nil {
		n, err  = client.buff.Read(b);
	}
	return n, err;
}

func (client *IRCClient) Write(b []byte) (n int, err os.Error) {
	var c IRCConn;

	if client.valid && client.ncon != nil {
		if client.tcon == nil {
			c = client.ncon;
		} else {
			c = client.tcon;
		}
		n, err = c.Write(b);
		if err != nil {
			client.Quit("connection failed");
		}
	}
	return n, err;
}

func (client *IRCClient) connect(name string, nick string, useTLS bool) (b bool, err os.Error) {
	var c net.Conn;

	// is this client already connected?
	if client.valid && client.ncon != nil {
		client.Quit("new connection");
	}

	c, err = net.Dial("tcp", "", name);
	if (err != nil) {
		fmt.Printf("error connecting to irc server %v, %v\n", name, err);
		client.ncon = nil;
		client.valid = false;
		return false, err;
	}
	if useTLS {
		tlsConfig := new(tls.Config);
		if tlsConfig != nil {
			file, err := os.Open("/dev/urandom", os.O_RDONLY, 0444);
			if file == nil || err != nil {
				c.Close();
				return false, err;
			}
			tlsConfig.Rand = file;
			tlsConfig.Time = time.Seconds;
			client.tcon = tls.Client(c, tlsConfig);
		} else {
			c.Close();
			return false, err;
		}
	} else {
		client.tcon = nil;
	}
	client.ncon = c;
	client.name = name;
	client.Nick = nick;
	client.valid = true;
	b = true;

	go _handle_server(client);

	return b, err;
}

func (client *IRCClient) Connect(name string, nick string) (b bool, err os.Error) {
	b, err = client.connect(name, nick, false);
	return b, err;
}

func (client *IRCClient) TLSConnect(name string, nick string) (b bool, err os.Error) {
	b, err = client.connect(name, nick, true);
	return b, err;
}

func (client *IRCClient) Quit(reason string) {
	if client.valid && client.ncon != nil {
		client.control <- false;
		<-client.control;
	}
	client.ncon = nil;
	client.tcon = nil;
	client.valid = false;
}

func (client *IRCClient) SetChannel(channel string) {
	if client != nil {
		client.Channel = channel;
	}
}

type IRCConn interface {
	Read(b []byte) (ret int, err os.Error);
	Write(b []byte) (ret int, err os.Error);
	Close() (err os.Error);
	SetReadTimeout(nsec int64) (err os.Error);
};

/* connect to a server, establish nick and user, respond to pings */
func _handle_server(client *IRCClient) {
	var c IRCConn;
	var _buf [4096]byte;

	if (client.tcon != nil) {
		c = client.tcon;
	} else {
		c = client.ncon;
	}
	nick := client.Nick;

	c.SetReadTimeout(1*1000*1000*1000); // 1 second timeout

	// sign in to the irc server
	n, err := c.Write(strings.Bytes("NICK :" + nick + "\r\n"));
	n, err = c.Write(strings.Bytes("USER " + nick + " mach mach :" + nick + "\r\n"));

	for {
		select {
		case <-client.control:
			c.Close();
			client.control <- false;
			return;
		default:
			n, err = c.Read(&_buf);
			if n > 0 {
				a := ping.MatchSlices(_buf[0:n-1]);
				if  a != nil {
					n, err = c.Write(strings.Bytes("PONG :"));
					n, err = c.Write(a[1]);
					n, err = c.Write(strings.Bytes("\r\n"));
				} else {
					/*
					for i := 0; i < n; i++ {
						fmt.Printf("%c", _buf[i]);
					}
					*/
					if client.buff == nil {
						client.buff = new(bytes.Buffer);
					}
					client.buff.Write(_buf[0:n]);
					client.CanRead <- true;
				}
			} else {
				check, ok := err.(*os.PathError);
				if ok && check.Error != os.EAGAIN {
					fmt.Printf("err: %v, %v\n", err, n);
					client.valid = false;
					c.Close();
					client.ncon = nil;
					client.tcon = nil;
					return;
				}
			}
		}
	}
}
