// Copyright 2009 Eric Estabrooks. All rights reserved.
// Use of this source code is governed by a BSD-style
// license that can be found in the LICENSE file.

// This package provides basic ansi terminal control
// including lightweight line editing with history
package term

import "os"
import "fmt"
import "bytes"
import "syscall"
import "strconv"
import "strings"
import "regexp"
import "container/vector"

import "./termios"

// Terminal information
type Term struct {
	row       int;
	col       int;
	ifd       int;
	ofd       int;
	hpos      int;
	history   *vector.StringVector;
	tios      *termios.Termios;
	original  *termios.Termios;
	raw       bool;
	valid     bool;
	MaxRow    int;
	MaxCol    int;
	write     chan *termWrite;
	wdone     chan bool;
}

type termWrite struct {
	row int;
	col int;
	s   string;
}

// Create a new terminal being fed from the given file descripor
// if fd is os.Stdin.Fd() then the terminal will use
// os.Stdout.Fd() for its outbound information
// otherwise it'll use fd for both directions
func NewTerm(fd int) (t *Term, err os.Error) {
	t = new(Term);

	err = t.SetFd(fd);
	t.Raw();
	t.Resize();

	t.history = new(vector.StringVector);

	return t, err;
}

// Return the terminals file descriptor
func (t* Term) Fd() int {
	return t.ifd;
}

// Set the terminals file descriptor
func (t* Term) SetFd(fd int) (err os.Error) {
	t.ifd = fd;
	if fd == os.Stdin.Fd() {
		t.ofd = os.Stdout.Fd();
	} else {
		t.ofd = fd;
	}

	if t.tios == nil {
		t.tios = new(termios.Termios);
	}

	if t.original == nil {
		t.original = new(termios.Termios);
	}

	if t.write == nil {
		t.write = make(chan *termWrite);
		go handleWrites(t);
	}

	if t.wdone == nil {
		t.wdone = make(chan bool);
	}

	err = t.original.Get(fd);
	if err == nil {
		t.tios.Copy(t.original);
		t.valid = true;
	} else {
		t.valid = false;
	}

	return err;
}

func handleWrites(t *Term) {
	//sbuf := []byte{033, '7'};
	//rbuf := []byte{033, '7'};

	for {
		item := <-t.write;
		//syscall.Write(t.ifd, sbuf);
		if item.row >= 0 {
			s := fmt.Sprintf("%c[%d;%dH", 033, item.row, item.col);
			syscall.Write(t.ifd, strings.Bytes(s));
		}
		syscall.Write(t.ifd, strings.Bytes(item.s));
		//syscall.Write(t.ifd, rbuf);
		t.wdone <- true;
	}
}

func (t *Term) Write(s string) {
	item := &termWrite{-1, -1, s};
	t.write <- item;
	<-t.wdone;
}

func (t *Term) WriteAt(row, col int, s string) {
	item := &termWrite{row, col, s};
	t.write <- item;
	<-t.wdone;
}

func (t *Term) WriteBytesAt(row, col int, b []byte) {
	conv := bytes.NewBuffer(b);
	item := &termWrite{row, col, conv.String()};
	t.write <- item;
	<-t.wdone;
}

// regex used to capture screen size
var ansiLoc *regexp.Regexp = regexp.MustCompile("(.*);(.*)R");

// check screen size and set MaxRow and MaxCol
// this should be called after an application receives a sigwinch
func (t* Term) Resize () os.Error {
	var buf [16]byte;

	if t.valid == false {
		return os.NewSyscallError("Term.Resize", int(syscall.EBADF))
	}

	t.MaxCol = 1;
	t.MaxRow = 1;

	save := []byte{033,'7'};
	fscroll := []byte{033, '[', 'r'};
	fquery := []byte{033, '[', '9', '9', '9', ';', '9', '9', '9', 'H', 033, '[', '6', 'n'};
	restore := []byte{033, '8'};

	if t.raw == false {
		tmp := termios.Dup(t.tios);
		tmp.LFlag = tmp.LFlag &^ termios.ICANON;
		err := tmp.Set(t.ifd);
		if err != nil {
			return err;
		}
	}

	syscall.Write(t.ofd, save);
	syscall.Write(t.ofd, fscroll);
	syscall.Write(t.ofd, fquery);
	n, _ := syscall.Read(t.ifd, &buf);
	syscall.Write(t.ofd, restore);

	if t.raw == false {
		err := t.tios.Set(t.ifd);
		if err != nil {
			return err;
		}
	}

	if n > 2 {
		conv := new(bytes.Buffer);
		conv.Write(buf[2:n]);

		res := ansiLoc.MatchStrings(conv.String());
		if res != nil && len(res) > 0 {
			t.MaxRow, _ = strconv.Atoi(res[1]);
			t.MaxCol, _ = strconv.Atoi(res[2]);
		}
	} else {
		return os.NewSyscallError("Term.Resize", int(syscall.EBADF));
	}

	return nil;
}

// put the terminal in pseudo raw mode, so we can do character grabs instead of line grabs
func (t *Term) Raw() (err os.Error) {
	if t.valid == false {
		return os.NewSyscallError("Term.Raw", int(syscall.EBADF));
	}

	t.tios.LFlag = t.tios.LFlag &^ ( termios.ICANON | termios.ECHO | termios.ECHONL );
	err = t.tios.Set(t.ifd);
	if err == nil {
		t.raw = true;
	} else {
		t.raw = false;
	}

	return err;
}

// restore the terminal to its original termios settings
func (t *Term) Restore() (err os.Error) {
	if t.valid == false {
		return os.NewSyscallError("Term.Restore", int(syscall.EBADF));
	}

	err = t.original.Set(t.ifd);
	if err == nil {
		t.raw = false;
	}

	return err;
}

func (t *Term) updatePosition() {
	var buf [16]byte;

	query := []byte{033, '[', '6', 'n'};

	syscall.Write(t.ofd, query);
	conv := new(bytes.Buffer);
	n, _ := syscall.Read(t.ifd, &buf);
	conv.Write(buf[2:n]);

	res := ansiLoc.MatchStrings(conv.String());
	if res != nil && len(res) > 0 {
		t.row, _ = strconv.Atoi(res[1]);
		t.col, _ = strconv.Atoi(res[2]);
	}
}

func (t *Term) clear(start, end int, b []byte) {
	for i:= start; i < end; i++ {
		b[i] = ' ';
	}
	t.WriteBytesAt(t.row, start, b[start:end]);
	t.CursorAt(t.row, start);
}

// Reads wherever the cursor currently is
func (t *Term) Read(b []byte) (n int, err os.Error) {
	t.updatePosition();
	n, err = t.ReadAt(t.row, t.col, b);

	return n, err;
}

// Reads in a line of text, provides basic editing and history
func (t *Term) ReadAt(row, col int, b []byte) (n int, err os.Error) {
	var _buf [1]byte;
	var nrow, ncol int;

	if t.valid == false {
		return 0, os.NewSyscallError("Term.Read", int(syscall.EBADF));
	}

	raw := t.raw;
	if raw == false {
		t.Raw();
	}

	t.row = row;
	t.col = col;

	t.CursorAt(row, col);

	n = 0; // number of characters in the buffer
	p := 0; // position in buffer
	overstrike := bool(false);
	escaped := bool(false);
	quotedinsert := bool(false);

	_, e := syscall.Read(t.ifd, &_buf);
	for e == 0 && _buf[0] != '\n' && _buf[0] != '\r' && n < len(b) {
		// insert the character or take other action
		if escaped == false && ((_buf[0] >= 32 && _buf[0] < 127) || quotedinsert) {
			//fmt.Printf("inserting %c\n", _buf[0]);
			if overstrike || p == n {
				b[p] = _buf[0];
				p++;
				if p > n {
					n = p;
				}
			} else { //insert mode
				for i := n; i > p; i-- {
					b[i] = b[i-1];
				}
				b[p] = _buf[0];
				p++;
				n++;
			}
			quotedinsert = false;
		} else { // check for editing character
			switch _buf[0] {
				case 1 : p = 0; // ctrl-A beginning of line
				case 2 : // move backward
					p--;
					if (p < 0) {
						p = 0;
					}
				case 4 : // delete forward
					if n > 1 {
						for i := p; i < (n-1); i++ {
							b[i] = b[i+1];
						}
						n--;
					}
				case 5 : p = n; // ctrl-E end of line
				case 6 : // move forward
					p++;
					if (p > n) {
						p = n;
					}
				case 11 : // delete line, from p to end
					t.clear(p, n, b);
					n = p;
				case 14 :
					n = t.historyFwd(b, n);
					p = 0;
				case 16 :
					n = t.historyBack(b, n);
					p = 0;
				case 18 : // ctrl-r i-search-back
					n = t.incSearchBack(b, n, row, col);
					p = 0;
				case 20 :
					if (n-p) > 1 {
						tmp := b[p];
						b[p] = b[p+1];
						b[p+1] = tmp;
					}
				case 21 : // delete line
					t.clear(0, n, b);
					n = 0;
					p = 0;
				case 22 : quotedinsert = true;
				case 27 : escaped = true;
				case 50 : overstrike = !overstrike;
				case 65 :
					n = t.historyBack(b, n);
					p = 0;
					escaped = false;
				case 66 :
					n = t.historyFwd(b, n);
					p = 0;
					escaped = false;
				case 67 : // move forward
					p++;
					if (p > n) {
						p = n;
					}
					escaped = false;
				case 68 : // move backward
					p--;
					if (p < 0) {
						p = 0;
					}
					escaped = false;
				case 91 : // anything?
				case 126 : escaped = false;
				case 127 : // delete backward
					if p > 0 {
						for i := p-1; i < (n-1); i++ {
							b[i] = b[i+1];
						}
						n--;
						p--;
					}
				default: fmt.Printf("unknown: %d\n", _buf[0]);
			}
		}
		// set cursor and redraw line from p (is line wrap going to be messy?)
		if p > 0 {
			t.writeLine(row, p+col, b[p-1:n]);
		} else {
			t.writeLine(row, col+1, b[0:n]);
		}
		if (p+col+1 > t.MaxCol) {
			nrow = (p+col)/t.MaxCol + row;
			ncol = (p+col+1) % t.MaxCol;
		} else {
			nrow = row;
			ncol = p+col+1;
		}
		t.CursorAt(nrow, ncol);
		_, e = syscall.Read(t.ifd, &_buf);
	}
	if (_buf[0] == '\n' || _buf[0] == '\r') && n < len(b) {
		b[n] = '\n';
		n++;
		//syscall.Write(t.ofd, strings.Bytes("\r\n"));
	}
	t.ClearFrom(row, col);
	t.ClearFrom(row+1, col);

	if raw == false {
		t.Restore();
	}

	return n, err;
}

/* writes out a line char by char, doing the ^ thing for ctrl chars, handles linewrap */
func (t *Term) writeLine(row, col int, b []byte) {
	var buf [1]byte;

	n := len(b);

	if col > t.MaxCol {
		trow := col/t.MaxCol;
		col = (col % t.MaxCol);
		row += trow;
		if col == 0 {
			col = t.MaxCol;
		}
	}

	for i := 0; i < n; i++ {
		buf[0] = b[i];
		if b[i] < ' ' {
			t.SaveCursor();
			s := fmt.Sprintf("%c[7m%c", 033, b[i] + 0x40);
			t.WriteAt(row, col, s);
			t.RestoreCursor();
		} else if b[i] > 126 {
			// what to do here, UTF-8 support?
		} else {
			t.WriteBytesAt(row, col, &buf);
		}
		col++;
		if col > t.MaxCol {
			col = 1;
			row++;
		}
	}
	buf[0] = ' ';
	t.WriteBytesAt(row, col, &buf);
}

// Move toward older entries in the history stack
func (t *Term) historyBack(b []byte, size int) int {
	var i int;

	hsize := t.history.Len();
	if hsize == 0 {
		return 0;
	}

	t.clear(0, size, b);

	t.hpos = (t.hpos + 1) % hsize;

	hb := strings.Bytes(t.history.At(t.hpos));

	for i = 0; i < len(hb) && i < len(b); i++ {
		b[i] = hb[i];
	}

	return i;
}

// Move toward newer entries in the history stack
func (t *Term) historyFwd(b []byte, size int) int {
	var i int;

	hsize := t.history.Len();
	if hsize == 0 {
		return 0;
	}

	t.clear(0, size, b);

	t.hpos--;
	if t.hpos < 0 {
		t.hpos = hsize - 1;
	}

	hb := strings.Bytes(t.history.At(t.hpos));

	for i = 0; i < len(hb) && i < len(b); i++ {
		b[i] = hb[i];
	}

	return i;
}

// search history for matches
func (t *Term) incSearchBack(b []byte, size, row, col int) int {
	var _buf   [1]byte;
	var ibuf   [256]byte;
	var hb     []byte;
	var i      int;
	var incPos int = 0;
	var match  bool = false;

	hsize := t.history.Len();
	if hsize == 0 {
		return 0;
	}

	// set up search prompt
	ibuf[0] = '?';
	n := int(1);

	// clear current and search line
	t.ClearFrom(row, col);
	t.ClearFrom(row+1, col);

	for {
		t.WriteBytesAt(row+1, col, ibuf[0:n]);
		_, e := syscall.Read(t.ifd, &_buf);
		if e == 0 {
			switch {
				case _buf[0] == 127:
					if (n > 1 ) {
						ibuf[n-1] = ' ';
						t.WriteBytesAt(row+1, col, ibuf[0:n]);
						n--;
						if (n == 1) {
							incPos = 0;
							t.ClearFrom(row, col);
							match = false;
						}
					} else {
						t.ClearFrom(row, col);
						t.ClearFrom(row+1, col);
						return size;
					}

				case _buf[0] == 18:
					incPos = (incPos + 1) % hsize;

				case _buf[0] == '\n' || _buf[0] == '\r':
					t.ClearFrom(row, col);
					t.ClearFrom(row+1, col);
					if match {
						for i = 0; i < len(hb) && i < len(b); i++ {
							b[i] = hb[i];
						}
						return i;
					} else {
						return size;
					}

				case _buf[0]>= ' ' && _buf[0] < 127: // not utf-8 compatible :(
					ibuf[n] = _buf[0];
					n++;
					if (n >= len(ibuf)) {
						n = len(ibuf)-1;
					}
					t.WriteBytesAt(row+1, col, ibuf[0:n]);
			}
			if n > 1 {
				hb = strings.Bytes(t.history.At(incPos));
				index := bytes.Index(hb, ibuf[1:n]);
				for index < 0 && incPos < (hsize-1) {
					incPos++;
					hb = strings.Bytes(t.history.At(incPos));
					index = bytes.Index(hb, ibuf[1:n])
				}
				if index >= 0 {
					t.ClearFrom(row, col);
					t.writeLine(row, col+1, hb);
					match = true;
				}
			}
		} else {
			break;
		}
	}

	return size;
}

// add to history and reset search position
func (t *Term) HistoryAdd(s string) {
	if t.history.Len() > 32 {
		t.history.Pop();
	}
	t.history.Insert(0, s);
	t.hpos = -1;
}

// move cursor to given location
func (t *Term) CursorAt(row, col int) {
	s := fmt.Sprintf("%c[%d;%dH", 033, row, col);
	t.Write(s);
}

// set text color/style/... (see ansi for numbers)
func (t *Term) Attribute(attribute int) {
	s := fmt.Sprintf("%c[%dm", 033, attribute);
	t.Write(s);
}

// clear the screen
func (t *Term) Clear() {
	s := fmt.Sprintf("%c[2J", 033);
	t.Write(s);
}

// clear current row
func (t *Term) ClearLine(row int) {
	t.SaveCursor();
	t.CursorAt(row, 0);
	s := fmt.Sprintf("%c[2K", 033);
	t.Write(s);
	t.RestoreCursor();
}

// clear from cursor to end of line
func (t *Term) ClearFrom(row, col int) {
	t.SaveCursor();
	t.CursorAt(row, col);
	s := fmt.Sprintf("%c[K", 033);
	t.Write(s);
	t.RestoreCursor();
}

// clear to cursor to end of line
func (t *Term) ClearTo(row, col int) {
	t.SaveCursor();
	t.CursorAt(row, col);
	s := fmt.Sprintf("%c[1K", 033);
	t.Write(s);
	t.RestoreCursor();
}

// move cursur up n rows
func (t *Term) CursorUp(rows int) {
	s := fmt.Sprintf("%c[%dA", 033, rows);
	t.Write(s);
}

// move cursur down n rows
func (t *Term) CursorDown(rows int) {
	s := fmt.Sprintf("%c[%dB", 033, rows);
	t.Write(s);
}

// move cursur left n cols
func (t *Term) CursorLeft(cols int) {
	s := fmt.Sprintf("%c[%dD", 033, cols);
	t.Write(s);
}

// move cursur right n cols
func (t *Term) CursorRight(cols int) {
	s := fmt.Sprintf("%c[%dC", 033, cols);
	t.Write(s);
}

// set scrolling area of screen, start of -1 sets fullscreen scroll
func (t *Term) SetScroll(start, end int) {
	var s string;
	if start == -1 {
		s = fmt.Sprintf("%c[r", 033);
	} else {
		s = fmt.Sprintf("%c[%d;%dr", 033, start, end);
	}
	t.Write(s);
}

// set scrolling area to full screen
func (t *Term) SetFullScroll() {
	s := fmt.Sprintf("%c[r", 033);
	t.Write(s);
}

// scroll the screen up by n rows
func (t *Term) ScrollUp(rows int) {
	s := fmt.Sprintf("%cM", 033);
	for i := 0; i < rows; i++ {
		t.Write(s);
	}
}

// scroll the screen down by n rows
func (t *Term) ScrollDown(rows int) {
	s := fmt.Sprintf("%cD", 033);
	for i := 0; i < rows; i++ {
		t.Write(s);
	}
}

func (t *Term) SaveCursor() {
	s := fmt.Sprintf("%c7", 033);
	t.Write(s);
}

func (t *Term) RestoreCursor() {
	s := fmt.Sprintf("%c8", 033);
	t.Write(s);
}

func (t *Term) LineWrap(on bool) {
	var s string;
	if on {
		s = fmt.Sprintf("%c[7h", 033);
	} else {
		s = fmt.Sprintf("%c[7l", 033);
	}
	t.Write(s);
}

// menu stuff, desctructive
