// Copyright 2009 Eric Estabrooks. All rights reserved.
// Use of this source code is governed by a BSD-style
// license that can be found in the LICENSE file.

// This package provides termios controls 
// Getting and Setting various termios attributes
package termios

// provides terminal control and some readline like editing for stdout/stdin

import "os"
import "syscall"
import "unsafe"

// ioctl constants
const (
	TCGETS = 0x5401;
	TCSETS = 0x5402;
)

// termios c_iflag constants
const (
	IGNBRK  = 0000001;
	BRKINT  = 0000002;
	IGNPAR  = 0000004;
	PARMRK  = 0000010;
	INPCK   = 0000020;
	ISTRIP  = 0000040;
	INLCR   = 0000100;
	IGNCR   = 0000200;
	ICRNL   = 0000400;
	IUCLC   = 0001000;
	IXON    = 0002000;
	IXANY   = 0004000;
	IXOFF   = 0010000;
	IMAXBEL = 0020000;
	IUTF8   = 0040000;
)

// termios c_oflag constants
const (
	OPOST  = 0000001;
	OLCUC  = 0000002;
	ONLCR  = 0000004;
	OCRNL  = 0000010;
	ONOCR  = 0000020;
	ONLRET = 0000040;
	OFILL  = 0000100;
	OFDEL  = 0000200;
)

// termios c_cflag (skipped for now
const (
	B0      = 0000000;
	B50     = 0000001;
	B75     = 0000002;
	B110    = 0000003;
	B134    = 0000004;
	B150    = 0000005;
	B200    = 0000006;
	B300    = 0000007;
	B600    = 0000010;
	B1200   = 0000011;
	B1800   = 0000012;
	B2400   = 0000013;
	B4800   = 0000014;
	B9600   = 0000015;
	B19200  = 0000016;
	B38400  = 0000017;
	B57600  = 0010001;
	B115200 = 0010002;
	B230400 = 0010003;
	B460800 = 0010004;
	B500000 = 0010005;
	B576000 = 0010006;
	B921600 = 0010007;
	B1000000= 0010010;
	B1152000= 0010011;
	B1500000= 0010012;
	B2000000= 0010013;
	B2500000= 0010014;
	B3000000= 0010015;
	B3500000= 0010016;
	B4000000= 0010017;

	CS5    = 0000000;
	CS6    = 0000020;
	CS7    = 0000040;
	CS8    = 0000060;
	CSIZE  = 0000060;
	CSTOPB = 0000100;
	CREAD  = 0000200;
	PARENB = 0000400;
	PARODD = 0001000;
	HUPCL  = 0002000;
	CLOCAL = 0004000;
)


// termios c_lflag constants
const (
	ISIG   = 0000001;
	ICANON = 0000002;
	ECHO   = 0000010;
	ECHOE  = 0000020;
	ECHOK  = 0000040;
	ECHONL = 0000100;
	NOFLSH = 0000200;
	TOSTOP = 0000400;
	IEXTEN = 0100000;
)

// control character constants
const (
	VINTR    = 0;
	VQUIT    = 1;
	VERASE   = 2;
	VKILL    = 3;
	VEOF     = 4;
	VTIME    = 5;
	VMIN     = 6;
	VSWTC    = 7;
	VSTART   = 8;
	VSTOP    = 9;
	VSUSP    = 10;
	VEOL     = 11;
	VREPRINT = 12;
	VDISCARD = 13;
	VWERASE  = 14;
	VLNEXT   = 15;
	VEOL2    = 16;
)

type Termios struct {
	IFlag  uint;
	OFlag  uint;
	CFlag  uint;
	LFlag  uint;
	Line   byte;
	CC     [32]byte;
	ISpeed uint;
	OSpeed uint;
}

// Gets the current termios attributes of the file descriptor
func (t *Termios) Get(fd int) (err os.Error) {
	p := unsafe.Pointer(t);
	_, _, e := syscall.Syscall(syscall.SYS_IOCTL, uintptr(fd), TCGETS, uintptr(p));

	if (e != 0) {
		return os.NewSyscallError("Termios.Get", int(e))
	}

	return nil;
}

// Sets the termios attributes of the file descriptor
func (t *Termios) Set(fd int) os.Error {
	p := unsafe.Pointer(t);
	_, _, e := syscall.Syscall(syscall.SYS_IOCTL, uintptr(fd), TCSETS, uintptr(p));

	if (e != 0) {
		return os.NewSyscallError("Termios.Set", int(e))
	}

	return nil;
}

// Copies from one Termios to another
func (dst *Termios) Copy(src *Termios) {
	if src == nil {
		return;
	}

	dst.IFlag = src.IFlag;
	dst.OFlag = src.OFlag;
	dst.CFlag = src.CFlag;
	dst.LFlag = src.LFlag;
	dst.Line = src.Line;
	dst.ISpeed = src.ISpeed;
	dst.OSpeed = src.OSpeed;

	for i := 0; i < 32; i++ {
		dst.CC[i] = src.CC[i];
	}
}

// Duplicate an existing Termios
func Dup(src *Termios) *Termios {
	dst := new(Termios);

	dst.Copy(src);

	return dst;
}

// Sets the file descriptor in char by char mode with most pre/post processing turned off
func (t *Termios) Raw(fd int) (err os.Error) {

	t.IFlag = t.IFlag &^ (IGNBRK | BRKINT | PARMRK | ISTRIP | INLCR | IGNCR | ICRNL | IXON);
	t.OFlag = t.OFlag &^ ( OPOST );
	t.LFlag = t.LFlag &^ ( ECHO | ECHONL | ICANON | ISIG | IEXTEN );
	t.CFlag = t.CFlag &^ ( CSIZE | PARENB );
	t.CFlag |= CS8;
	t.CC[VMIN] = 1;
	t.CC[VTIME] = 0;

	err = t.Set(fd);

	return err;
}

// Sets the file descriptor in a line by line mode with signal support
func (t *Termios) Cooked(fd int) (err os.Error) {

	t.IFlag |=  ( IGNBRK | BRKINT | PARMRK | ISTRIP | INLCR | IGNCR | ICRNL | IXON );
	t.OFlag |= ( OPOST );
	t.LFlag |= ( ECHO | ECHONL | ICANON | ISIG | IEXTEN );
	t.CFlag = t.CFlag &^ ( CSIZE | PARENB );
	t.CFlag |= CS8;
	t.CC[VMIN] = 0;
	t.CC[VTIME] = 0;

	err = t.Set(fd);

	return err;
}
